# -------------------------------------------------------------------------------- NOTEBOOK-CELL: CODE
# -*- coding: utf-8 -*-
import os
import tempfile
import dataiku
from datasets import Dataset
import torch
from transformers import AutoModelForCausalLM, AutoTokenizer, BitsAndBytesConfig, TrainingArguments
from trl import SFTTrainer, DataCollatorForCompletionOnlyLM
from peft import LoraConfig

MODEL_NAME = "mistralai/Mistral-7B-Instruct-v0.2"
MODEL_REVISION = "c72e5d1908b1e2929ec8fc4c8820e9706af1f80f"

train_dataset = Dataset.from_pandas(
    dataiku.Dataset("sft_train").get_dataframe()
)
validation_dataset = Dataset.from_pandas(
    dataiku.Dataset("sft_validation").get_dataframe()
)

folder = dataiku.Folder("74GhsK3Z")

# -------------------------------------------------------------------------------- NOTEBOOK-CELL: CODE
# Code inspired by the TRL documentation: https://huggingface.co/docs/trl/main/en/sft_trainer

bnb_config = BitsAndBytesConfig(
    load_in_4bit=True,
    bnb_4bit_quant_type="nf4",
    bnb_4bit_compute_dtype=torch.float16,
)

model = AutoModelForCausalLM.from_pretrained(
    MODEL_NAME,
    revision=MODEL_REVISION,
    device_map="auto",
    quantization_config=bnb_config,
)
model.config.use_cache = False # Because the model will change as it is fine-tuned

tokenizer = AutoTokenizer.from_pretrained(
    MODEL_NAME,
    revision=MODEL_REVISION
)
tokenizer.pad_token = tokenizer.eos_token

# -------------------------------------------------------------------------------- NOTEBOOK-CELL: CODE
def formatting_prompts_func(example):
    """
    Convert a batch of training examples into the model's instruction format.
    """
    output_texts = []
    for i in range(len(example['input'])):
        text = f"[INST] Extract key details from: {example['input'][i]} [/INST] {example['output'][i]}"
        output_texts.append(text)
    return output_texts

response_template = tokenizer.encode("[/INST]", add_special_tokens=False)
collator = DataCollatorForCompletionOnlyLM(response_template, tokenizer=tokenizer)

with tempfile.TemporaryDirectory() as temp_dir:
    
    # Define the training parameters
    training_arguments = TrainingArguments(
        per_device_train_batch_size=16,
        optim="paged_adamw_8bit",
        gradient_checkpointing=True,
        output_dir=temp_dir
    )

    peft_config = LoraConfig(
        r=16,
        lora_alpha=32,
        lora_dropout=0.05,
        task_type="CAUSAL_LM",
    )
    
    trainer = SFTTrainer(
        model,
        args=training_arguments,
        formatting_func=formatting_prompts_func,
        data_collator=collator,
        train_dataset=train_dataset,
        eval_dataset=validation_dataset,
        peft_config=peft_config,
        max_seq_length=512,
        neftune_noise_alpha=5
    )
    
    # Fine-tune the model
    trainer.train()
    
    # Save the fine-tuned model in the managed folder
    trainer.save_model(output_dir=temp_dir)
    for root, dirs, files in os.walk(temp_dir):
        for file in files:
            source_path = os.path.join(root, file)
            target_path = os.path.relpath(source_path, temp_dir)
            folder.upload_file(target_path, source_path)