# -------------------------------------------------------------------------------- NOTEBOOK-CELL: CODE
import dataiku
from dataiku.langchain.dku_llm import DKULLM
import json
import mlflow
from spacy import displacy

from langchain.agents import AgentExecutor, create_react_agent

from langchain_core.prompts import PromptTemplate
from langchain_core.prompts import MessagesPlaceholder
from langchain.tools import StructuredTool
from langchain.callbacks import MlflowCallbackHandler

from fictitious_tools import (
    get_customer_id,
    get_details,
    reset_password,
    cancel_appointment,
    schedule_local_intervention,
    schedule_distant_intervention,
    run_diagnostics,
    sign_up_to_option,
    cancel_option
)

df = dataiku.Dataset("requests").get_dataframe()

LLM_ID = dataiku.get_custom_variables()["LLM_instruct_id"]

project = dataiku.api_client().get_default_project()
mlflow_handle = project.setup_mlflow(project.get_managed_folder("EnynTX9x"))
EXPERIMENT_NAME = "agent"
mlflow.set_experiment(EXPERIMENT_NAME)

# -------------------------------------------------------------------------------- NOTEBOOK-CELL: CODE
functions = [
    get_details,
    reset_password,
    cancel_appointment,
    schedule_local_intervention,
    schedule_distant_intervention,
    run_diagnostics,
    sign_up_to_option,
    cancel_option,
    get_customer_id
]

def get_partial_functions(customer_id):
    """
    Provide partial functions corresponding to a specific customer.
    These partial functions should be used when the customer is identified.
    """
    return [
        lambda s="": get_details(customer_id),
        lambda s="": reset_password(customer_id),
        lambda s="": cancel_appointment(customer_id),
        lambda s="": schedule_local_intervention(customer_id),
        lambda s="": schedule_distant_intervention(customer_id),
        lambda s="": run_diagnostics(customer_id),
        lambda s="": sign_up_to_option(f"{customer_id},{s}"),
        lambda s="": cancel_option(f"{customer_id},{s}"),
    ]

def get_tools(customer_id=None):
    """
    Provide either the general tools or the customer-specific tools.
    """
    tools = []
    if customer_id is None:
        for f in functions:
            tools.append(StructuredTool.from_function(f))
    else:
        partial_functions = get_partial_functions(str(customer_id))
        for i in range(len(partial_functions)):
            description = "\n".join(functions[i].__doc__.strip().split("\n    ")[:-1])
            # If partial functions are used, their description should be amended
            if functions[i].__name__ in ["sign_up_to_option", "cancel_option"]:
                description += "\nThe action input should be either 'TV' or 'Premium'"
            else:
                description += "\nThe action input should be the empty string"
            tools.append(StructuredTool.from_function(
                partial_functions[i],
                name=functions[i].__name__,
                description=description
            ))
    return tools

# -------------------------------------------------------------------------------- NOTEBOOK-CELL: CODE
prompt = PromptTemplate(
    input_variables=["tools", "tool_names", "chat_history", "agent_scratchpad"],
    template="""Assistant is a large language model trained by OpenAI.

Assistant is designed to be able to assist with a wide range of tasks, from answering simple questions to providing in-depth explanations and discussions on a wide range of topics. As a language model, Assistant is able to generate human-like text based on the input it receives, allowing it to engage in natural-sounding conversations and provide responses that are coherent and relevant to the topic at hand.

Assistant is constantly learning and improving, and its capabilities are constantly evolving. It is able to process and understand large amounts of text, and can use this knowledge to provide accurate and informative responses to a wide range of questions. Additionally, Assistant is able to generate its own text based on the input it receives, allowing it to engage in discussions and provide explanations and descriptions on a wide range of topics.

Overall, Assistant is a powerful tool that can help with a wide range of tasks and provide valuable insights and information on a wide range of topics. Whether you need help with a specific question or just want to have a conversation about a particular topic, Assistant is here to assist.

TOOLS:
------

Assistant has access to the following tools:

{tools}

To use a tool, please use the following format:

```
Thought: Do I need to use a tool? Yes
Action: the action to take, should be one of [{tool_names}]
Action Input: the input to the action
Observation: the result of the action
```

When you have a response to say to the Human, or if you do not need to use a tool, you MUST use the format:

```
Thought: Do I need to use a tool? No
Final Answer: [your response here]
```

Begin!

Previous conversation history:
{chat_history}

New input: {input}
{agent_scratchpad}"""
)

def serialize_intermediate_steps(l):
    """
    Serialize the intermediate steps returned by the agent.
    """
    return json.dumps([(x[0].tool, x[0].tool_input, x[0].log, x[1]) for x in l])

def process_request(request, customer_id=None):
    """
    Handle the customer's request.
    If the customer id is specified, only the corresponding customer can be affected by the agent.
    """
    mlflow_callback = MlflowCallbackHandler(
        experiment=EXPERIMENT_NAME,
        tags={"model": LLM_ID}
    )

    llm = DKULLM(
        llm_id=LLM_ID,
        temperature=0,
        callbacks=[mlflow_callback]
    )

    if customer_id is not None:
        identity = get_details(str(customer_id)).split(",")[0]
        request = f"{request} (message received from {identity})"

    tools = get_tools(customer_id=customer_id)
    agent = create_react_agent(llm, tools, prompt)
    agent_executor = AgentExecutor(
        agent=agent,
        tools=tools,
        verbose=True,
        return_intermediate_steps=True,
        handle_parsing_errors=True,
        callbacks=[mlflow_callback]
    )

    result = agent_executor.invoke({"input": request, "chat_history": []})

    # List the actions and their details in two strings returned along with the answer
    actions = ""
    for j in range(len(result["intermediate_steps"])):
        tool = result["intermediate_steps"][j][0].tool
        tool_input = result["intermediate_steps"][j][0].tool_input
        thought = result["intermediate_steps"][j][0].log
        observation = result["intermediate_steps"][j][1]
        actions += f"{j+1}. {tool}({tool_input})\n"

    return (
        result["output"],
        actions[:-1] if len(actions) > 0 else actions,
        serialize_intermediate_steps(result["intermediate_steps"])
    )

# -------------------------------------------------------------------------------- NOTEBOOK-CELL: CODE
for i in range(len(df)):
    reply, actions, intermediate_steps = process_request(
        df.iloc[i].request,
        customer_id=df.iloc[i].customer_id
    )
    df.at[i, "draft_reply"] = reply
    df.at[i, "actions"] = actions
    df.at[i, "intermediate_steps"] = intermediate_steps

# -------------------------------------------------------------------------------- NOTEBOOK-CELL: CODE
dataiku.Dataset("requests_processed").write_with_schema(df)