import os
import re
import time
from functools import partial, lru_cache
import dataiku
from dataiku.langchain.dku_llm import DKUChatLLM

from datetime import datetime
import json

from langchain.agents import AgentExecutor, create_react_agent
from langchain_core.prompts import PromptTemplate
from langchain_core.prompts import MessagesPlaceholder
from langchain.tools import StructuredTool

import dash
from dash import dcc
from dash import html
from dash.dependencies import Input, Output, State
import dash_bootstrap_components as dbc

from fictitious_tools import (
    get_customer_id,
    get_details,
    reset_password,
    cancel_appointment,
    schedule_local_intervention,
    schedule_distant_intervention,
    run_diagnostics,
    sign_up_to_option,
    cancel_option
)

LOG_ALL_ANSWERS = False # Whether all answers or only answers with user feedback should be logged
WEBAPP_NAME = "agent_webapp" # Name of the app (logged when an answer is flagged)
VERSION = dataiku.get_custom_variables()[WEBAPP_NAME] # Version of the app (logged when an answer is flagged)
LLM_ID = dataiku.get_custom_variables()["LLM_instruct_id"]

# Folder to log answers and positive/negative reactions
FOLDER = dataiku.Folder("zhYE6r2h")

df = dataiku.Dataset("requests").get_dataframe()
customers_df = dataiku.Dataset("customers").get_dataframe()

if len(LLM_ID) > 0:
    llm_provided = True
    llm = dataiku.api_client().get_default_project().get_llm(LLM_ID)
else:
    llm_provided = False

ERROR_MESSAGE_MISSING_KEY = """
LLM Connection missing. You need to add it as a user secret. Cf. this project's wiki

**Please note that this web app is not live on Dataiku’s public project gallery but you can test it by downloading the project and providing an LLM connection**.

You can find examples of answers in the `requests_processed` dataset.
"""

#####################################################################
### Modify the code below to adapt this example to your own tools ###
#####################################################################

functions = [
    get_details,
    reset_password,
    cancel_appointment,
    schedule_local_intervention,
    schedule_distant_intervention,
    run_diagnostics,
    sign_up_to_option,
    cancel_option,
    get_customer_id
]

def get_partial_functions(customer_id):
    return [
        lambda s: get_details(customer_id),
        lambda s: reset_password(customer_id),
        lambda s: cancel_appointment(customer_id),
        lambda s: schedule_local_intervention(customer_id),
        lambda s: schedule_distant_intervention(customer_id),
        lambda s: run_diagnostics(customer_id),
        lambda s: sign_up_to_option(f"{customer_id},{s}"),
        lambda s: cancel_option(f"{customer_id},{s}"),
    ]

def get_tools(customer_id=None):
    tools = []
    if customer_id is None:
        for f in functions:
            tools.append(StructuredTool.from_function(f))
    else:
        partial_functions = get_partial_functions(str(customer_id))
        for i in range(len(partial_functions)):
            description = "\n".join(functions[i].__doc__.strip().split("\n    ")[:-1])
            if functions[i].__name__ in ["sign_up_to_option", "cancel_option"]:
                description += "\nThe action input should be either 'TV' or 'Premium'"
            else:
                description += "\nThe action input should be the empty string"
            tools.append(StructuredTool.from_function(
                partial_functions[i],
                name=functions[i].__name__,
                description=description
            ))
    return tools

#####################################################################
### Modify the code above to adapt this example to your own tools ###
#####################################################################

prompt = PromptTemplate(
    input_variables=["tools", "tool_names", "chat_history", "agent_scratchpad"],
    template="""Assistant is a large language model trained by OpenAI.

Assistant is designed to be able to assist with a wide range of tasks, from answering simple questions to providing in-depth explanations and discussions on a wide range of topics. As a language model, Assistant is able to generate human-like text based on the input it receives, allowing it to engage in natural-sounding conversations and provide responses that are coherent and relevant to the topic at hand.

Assistant is constantly learning and improving, and its capabilities are constantly evolving. It is able to process and understand large amounts of text, and can use this knowledge to provide accurate and informative responses to a wide range of questions. Additionally, Assistant is able to generate its own text based on the input it receives, allowing it to engage in discussions and provide explanations and descriptions on a wide range of topics.

Overall, Assistant is a powerful tool that can help with a wide range of tasks and provide valuable insights and information on a wide range of topics. Whether you need help with a specific question or just want to have a conversation about a particular topic, Assistant is here to assist.

TOOLS:
------

Assistant has access to the following tools:

{tools}

To use a tool, please use the following format:

```
Thought: Do I need to use a tool? Yes
Action: the action to take, should be one of [{tool_names}]
Action Input: the input to the action
Observation: the result of the action
```

When you have a response to say to the Human, or if you do not need to use a tool, you MUST use the format:

```
Thought: Do I need to use a tool? No
Final Answer: [your response here]
```

Begin!

Previous conversation history:
{chat_history}

New input: {input}
{agent_scratchpad}"""
)

def serialize_intermediate_steps(l):
    """
    Serialize the intermediate steps returned by the agent.
    """
    return json.dumps([(x[0].tool, x[0].tool_input, x[0].log, x[1]) for x in l])

def process_request(request, customer_id=None):
    """
    Handle the customer's request.
    If the customer id is specified, only the corresponding customer can be affected by the agent.
    """

    llm = DKUChatLLM(
        llm_id=LLM_ID,
        temperature=0
    )

    if customer_id is not None:
        identity = get_details(str(customer_id)).split(",")[0]
        request = f"{request} (message received from {identity})"

    tools = get_tools(customer_id=customer_id)
    agent = create_react_agent(llm, tools, prompt)
    agent_executor = AgentExecutor(
        agent=agent,
        tools=tools,
        verbose=True,
        return_intermediate_steps=True,
        handle_parsing_errors=True
    )

    result = agent_executor.invoke({"input": request, "chat_history": []})

    # List the actions and their details in two strings returned along with the answer
    actions = ""
    for j in range(len(result["intermediate_steps"])):
        tool = result["intermediate_steps"][j][0].tool
        tool_input = result["intermediate_steps"][j][0].tool_input
        thought = result["intermediate_steps"][j][0].log
        observation = result["intermediate_steps"][j][1]
        actions += f"{j+1}. {tool}({tool_input})\n"

    return (
        result["output"],
        actions[:-1] if len(actions) > 0 else actions,
        serialize_intermediate_steps(result["intermediate_steps"])
    )

def format_actions(d):
    """
    Format the actions taken by the agent.
    """
    if len(d) == 0:
        return ""
    return "**Actions taken**\n\n" + "\n\n".join([f"{i+1}. **{d[i][0]}({d[i][1]})**: {d[i][3]}" for i in range(len(d))])

# Layout

STYLE_ANSWER = {
    "margin-top": "20px",
    "align-items": "flex-start",
    "display": "flex",
    "height": "auto"
}

STYLE_BUTTON = {
    "width": "20px",  
    "text-align": "center",
    "margin": "0px 5px",
}

STYLE_FEEDBACK = {
    "margin-left": "10px",
    "display": "none"
}

STYLE_STEPS = {
    "white-space": "pre-line"
}

ok_icon_fill = html.Span(html.I(className="bi bi-emoji-smile-fill"))
nok_icon_fill = html.Span(html.I(className="bi bi-emoji-frown-fill"))
ok_icon = html.Span(html.I(className="bi bi-emoji-smile"))
nok_icon = html.Span(html.I(className="bi bi-emoji-frown"))

send_icon = html.Span(html.I(className="bi bi-send"))
question_bar = dbc.InputGroup(
    [
        dbc.Input(id='query', value='', type='text', minLength=0),
        dbc.Button(send_icon, id='send-btn', title='Get an answer')
    ],
    style = {"margin-top": "20px"}
)

app.title = "Question answering"
app.config.external_stylesheets = [
    dbc.themes.ZEPHYR,
    dbc.icons.BOOTSTRAP
]

options_context = [
    {
        "label": f"Customer #{customers_df.iloc[i].id} ({customers_df.iloc[i].title} {customers_df.iloc[i].first_name} {customers_df.iloc[i].last_name})",
        "value": customers_df.iloc[i].id
    }
    for i in range(len(customers_df))
] + [{"label": "All customers", "value": -1}]
options_display = [
    {"label": "Reply only", "value": 0},
    {"label": "Reply and intermediate steps", "value": 1},
]

app.layout = html.Div(
    [
        html.H4(
            "LLM agent",
            style={"margin-top": "20px", "text-align": "center"}
        ),
        dbc.Form(
            dbc.Row(
                [
                    dbc.Label("Context", width="auto"),
                    dbc.Col(dbc.Select(
                        id='customer_id',
                        options=options_context,
                        value=1,
                    ), className="me-3"),
                    dbc.Label("Display", width="auto"),
                    dbc.Col(dbc.Select(
                        id='display_mode',
                        options=options_display,
                        value=1,
                    ), className="me-3"),
                ],
                className="g-2"
            ),
            style = {"margin-top": "20px"}
        ),
        question_bar,   
        html.Div(
            [
                dbc.Spinner(
                    dcc.Markdown(
                        id='answer',
                        link_target="_blank"
                    ),
                    color="primary",
                    spinner_style={"display": "none"}
                ),
                html.Div(
                    [
                        html.A(ok_icon, id="link_ok", href="#", style=STYLE_BUTTON),
                        html.A(nok_icon, id="link_nok", href="#", style=STYLE_BUTTON)   
                    ],
                    id="feedback_buttons",
                    style=STYLE_FEEDBACK
                )
            ],
            style=STYLE_ANSWER
        ),

        dbc.Spinner(
            dcc.Markdown(
                id='intermediate_steps',
                link_target="_blank",
                style=STYLE_STEPS
            ),
            color="primary"
        ),
        html.Div(id='debug'),
        dcc.Store(id='feedback', data=2, storage_type='memory'),
        dcc.Store(id='question', storage_type='memory'),
        dcc.Store(id='question_id', storage_type='memory'),
    ],
    style={
        "margin": "auto",
        "text-align": "left",
        "max-width": "800px"
    }
)

# Callbacks

@app.callback(
    Output('answer', 'children'),
    Output('intermediate_steps', 'children'),
    Output('feedback_buttons', 'style'),
    Output('question', 'data'),
    Output('question_id', 'data'),
    Input('send-btn', 'n_clicks'),
    Input('query', 'n_submit'),
    State('customer_id', 'value'),
    State('query', 'value'),
)
def answer_question(n_clicks, n_submit, customer_id, query):
    """
    Display the answer
    """
    if len(query) == 0:
        return "", "", STYLE_FEEDBACK, query, 0
    start = time.time()
    if not llm_provided:
        return "", ERROR_MESSAGE_MISSING_KEY, STYLE_FEEDBACK, query, 0
    style = dict(STYLE_FEEDBACK)
    style["display"] = "flex"
    customer_id = None if customer_id == "-1" else customer_id
    answer, _, actions_details = process_request(
        query,
        customer_id=customer_id
    )
    answer = answer.replace("[", "\[").replace("]", "\]")
    resources = f"\n\n{(time.time()-start):.1f} seconds"
    return f"** {answer} **{resources}", format_actions(json.loads(actions_details)), style, query, hash(str(start) + query)

@app.callback(
    Output('intermediate_steps', 'style'),
    Input('display_mode', 'value'),
    prevent_initial_call=True
)
def display_intermediate_steps(mode):
    """
    Toggle the display of the intermediate steps
    """
    style = dict(STYLE_STEPS)
    style["display"] = "block" if mode == "1" else "none"
    return style

@app.callback(
    Output('debug', 'children'),
    Input('answer', 'children'),
    Input('link_ok', 'children'),
    Input('link_nok', 'children'),
    Input('intermediate_steps', 'children'),
    State('query', 'value'),
    State('question_id', 'data'),
    State('feedback', 'data'),
)
def log_answer(answer, ok, nok, intermediate_steps, query, question_id, feedback):
    """
    Log the question and the answer
    """
    if llm_provided and len(answer) > 0:
        path = f"/{str(question_id)}.json"
        if LOG_ALL_ANSWERS or feedback in [1, -1]:
            with FOLDER.get_writer(path) as w:
                w.write(bytes(json.dumps({
                    "question": query,
                    "answer": f"Reply: {answer}\n\nIntermediate steps: {intermediate_steps}",
                    "feedback": 0 if feedback == 2 else feedback,
                    "timestamp": str(datetime.now()),
                    "version": VERSION,
                    "webapp": WEBAPP_NAME
                }), "utf-8"))
        else:
            if path in FOLDER.list_paths_in_partition():
                FOLDER.delete_path(path)
    return ""

@app.callback(
    Output('link_ok', 'children'),
    Output('link_nok', 'children'),
    Input('feedback', 'data')
)
def update_icons(value):
    """
    Update the feedback icons when the user likes or dislikes an answer
    """
    ok = ok_icon_fill if value is not None and value == 1 else ok_icon
    nok = nok_icon_fill if value is not None and value == -1 else nok_icon
    return ok, nok

@app.callback(
    Output('feedback', 'data'),
    Input('link_ok', 'n_clicks'),
    Input('link_nok', 'n_clicks'),
    Input('send-btn', 'n_clicks'),
    Input('query', 'n_submit'),
    State('feedback', 'data'),
    State('query', 'value'),
    State('question', 'data'),
)
def provide_feedback(ts_ok, ts_nok, click, submit, value, question, previous_question):
    """
    Record the feedback of the user
    """
    triggered = dash.ctx.triggered_id
    if triggered == "link_ok":
        return 1 if value != 1 else 0
    elif triggered == "link_nok":
        return -1 if value != -1 else 0
    return value if question == previous_question else 2